{*******************************************************}
{                                                       }
{              Delphi VCL Extensions                    }
{                                                       }
{          Copyright (c) 1999 Andrei Roofin             }
{                                                       }
{          www - http://dsr.freeservers.com             }
{             mail - ray@mail.tomsknet.ru               }
{                                                       }
{*******************************************************}
unit rDragController;

{$IFDEF VER80} {$DEFINE D4LESS} {$ENDIF}
{$IFDEF VER90} {$DEFINE D4LESS} {$ENDIF}
{$IFDEF VER95} {$DEFINE D4LESS} {$ENDIF}
{$IFDEF VER100} {$DEFINE D4LESS} {$ENDIF}
{$IFDEF VER110} {$DEFINE D4LESS} {$ENDIF}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  stdCtrls, checklst, commctrl;

type
  TrMovedEvent = procedure (Sender: TObject; FromIndex, ToIndex: Longint) of object;


  TrDragController = class(TComponent)
  private
    FDragRectIdx: integer;
    DragMsgID : cardinal;
    FListBox: TCustomListBox;
    oldLBWnd: TWndMethod;
    oldParentWnd: TWndMethod;
    FOnItemMove: TrMovedEvent;
    procedure SetListBox(const Value: TCustomListBox);
  protected
    procedure DragRect(lb:TCustomListBox; const idx:integer);
    procedure EndDrag(lb:TCustomListBox);
    procedure DragOver(lb:TCustomListBox; const Point:TPoint);
    procedure DragDrop(lb:TCustomListBox; const Point:TPoint);
    function isCheckListBox(lb:TCustomListBox):boolean;
    function isCheckPoint(lb:TCustomListBox; Point:TPoint): boolean;
    procedure PrepareDrag(LBShow:boolean);
    procedure newLBWnd(var Message: TMessage);
    procedure newParentWnd(var Message: TMessage);
  public
    procedure ItemMove(lb:TCustomListBox; const src,dest:integer); virtual;
    destructor Destroy; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  published
    property ListBox: TCustomListBox read FListBox write SetListBox;
    property OnItemMove : TrMovedEvent read FOnItemMove write FOnItemMove;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Ray', [TrDragController]);
end;


Type
  THackCheckListBox = class(TCheckListBox);


var
  MyDragMsg: cardinal;


{ TrDragController }


destructor TrDragController.Destroy;
begin
 ListBox := nil;
 inherited Destroy;
end;

procedure TrDragController.newLBWnd(var Message: TMessage);
begin
 if Message.Msg=CM_SHOWINGCHANGED then PrepareDrag(ListBox.Showing);
 oldLBWnd(Message);
end;


procedure TrDragController.newParentWnd(var Message: TMessage);
var
 p: PDragListInfo;
 i: Integer;
 lb: TCustomListbox;
begin
 if message.Msg=MyDragMsg then begin
  Message.Result := MyDragMsg;
 end else
 if message.Msg=DragMsgID then begin
   p := PDragListInfo(message.lParam);
   lb:= FindControl(p^.hWnd) as TCustomListBox;
   case p^.uNotification of
    DL_BEGINDRAG : begin
                    FDragRectIdx := -1;
                    if isCheckListBox(lb) AND
                       isCheckPoint(lb, lb.ScreenToClient(p^.ptCursor))
                     then begin
                      i := LBItemFromPt(p^.hWnd, p^.ptCursor,false);
                      with THackCheckListBox(lb) do begin
                       Checked[i] := not Checked[i];
                       ClickCheck;
                      end;    // with
                      message.Result:=0;
                     end else begin
                      message.Result:=1;
                      lb.Cursor := crDrag;
                     end;
                   end;
    DL_CANCELDRAG : begin
                     EndDrag(lb);
                    end;
    DL_DRAGGING : begin
                     DragOver(lb,p^.ptCursor);
                     Message.Result := 0;
                  end;
    DL_DROPPED : begin
                  if FDragRectIdx>=0 then begin
                   EndDrag(lb);
                   DragDrop(lb, p^.ptCursor);
                  end else lb.Cursor := crDefault;
                 end;
   end;
 end else
  oldParentWnd(message);
end;


procedure TrDragController.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
 inherited Notification(AComponent,Operation);
 if (AComponent=FListBox) and (Operation=opRemove)
  then ListBox := nil;

end;

procedure TrDragController.SetListBox(const Value: TCustomListBox);
begin
 if FListBox<>Value then begin
  if (Value<>nil) and THackCheckListBox(Value).MultiSelect then begin
   Application.MessageBox('Sorry, can''t operate with ListBox where MultiSelect=True','TrDragController Error',MB_Ok);
   Exit;
  end;

  if FListBox<>nil then begin
   PrepareDrag(False);
   if not (csDesigning in ComponentState) then FListBox.WindowProc := oldLBWnd;
  end;
  FListBox := Value;
  if FListBox<>nil then begin
   if not (csDesigning in ComponentState) then begin
    oldLBWnd := FListBox.WindowProc;
    FListBox.WindowProc := newLBWnd;
   end;
  end;
 end;
end;



procedure TrDragController.PrepareDrag(LBShow: boolean);
begin
 if not (csDesigning in ComponentState) then  begin
   if LBShow then begin
    MakeDragList(ListBox.Handle);
    if SendMessage(ListBox.Parent.handle,MyDragMsg,0,0)<>integer(MyDragMsg) then begin
     oldParentWnd := ListBox.Parent.WindowProc;
     ListBox.Parent.WindowProc := newParentWnd;
     DragMsgID := RegisterWindowMessage(DRAGLISTMSGSTRING);
    end;
   end else begin
    if (ListBox<>nil) and (ListBox.Parent<>nil)
     then ListBox.Parent.WindowProc := oldParentWnd;
   end;
 end;
end;

function TrDragController.isCheckListBox(lb:TCustomListBox): boolean;
begin
 Result := (lb<>nil) and (lb is TCheckListBox);
end;


function TrDragController.isCheckPoint(lb:TCustomListBox; Point: TPoint): boolean;
var
 i: integer;
begin
if lb is TCheckListBox then
 with THackCheckListBox(lb) do begin
  i := ItemAtPos(Point,False);
  Result := False;
  if i>=0 then
{$IFNDEF D4LESS}
    if not UseRightToLeftAlignment then begin
{$ENDIF}
     if point.X - ItemRect(i).Left < GetCheckWidth
      then Result := True;
{$IFNDEF D4LESS}
    end else begin
      Dec(point.x, ItemRect(i).Right - GetCheckWidth);
      if (point.x > 0) and (point.x < GetCheckWidth)
       then Result := True;
    end;
{$ENDIF}
 end
 else Result := False;   
end;



procedure TrDragController.DragRect(lb:TCustomListBox; const idx: integer);
var c:TColor;
begin
  with lb,Canvas do begin
   c := Brush.Color;
   Brush.Color := clYellow;
   DrawFocusRect(ItemRect(idx));
   Brush.Color := c;
  end;    // with
end;


procedure TrDragController.EndDrag(lb:TCustomListBox);
begin
 if (FDragRectIdx>=0) and
    (FDragRectIdx<lb.Items.Count) and
    (FDragRectIdx<>lb.ItemIndex)
  then DragRect(lb,FDragRectIdx);
 lb.Cursor := crDefault;
 FDragRectIdx := -1;
end;

procedure TrDragController.DragDrop(lb:TCustomListBox; const Point: TPoint);
var
 i:integer;
begin
  i:=LBItemFromPt(lb.Handle,point,False);
  if i=lb.Items.Count then i:=i-1;
  ItemMove(lb,lb.ItemIndex,i);
  lb.ItemIndex := i;
end;

procedure TrDragController.DragOver(lb:TCustomListBox; const Point: TPoint);
var
 i: integer;
begin
 with lb do begin
   i := ItemAtPos(ScreenToClient(Point),False);
//   if i<0 then Screen.Cursor := crNoDrop
//    else Screen.Cursor := crDefault;

   if i=FDragRectIdx then i:=LBItemFromPt(handle,point,true);
   if i<>FDragRectIdx then begin

    if (FDragRectIdx>-1) and (FDragRectIdx<>ItemIndex) then DragRect(lb,FDragRectIdx);
     i := LBItemFromPt(handle,point,true);
     FDragRectIdx := i;
    if (FDragRectIdx>-1) and (FDragRectIdx<>ItemIndex) then DragRect(lb,FDragRectIdx);

   end;
 end;
end;


procedure TrDragController.ItemMove(lb:TCustomListBox; const src, dest: integer);
var c:boolean;
begin
 with lb do begin
  if (Src<0) OR (Src>=Items.Count) then Exit;
  if (Dest<0) OR (Dest>=Items.Count) then Exit;

  if isCheckListBox(lb) then begin
   c := TChecklistBox(lb).Checked[src];
   Items.Move(src,dest);
   TChecklistBox(lb).Checked[dest] := c;
  end else Items.Move(src,dest);
 end;
 if Assigned(FOnItemMove) then FOnItemMove(lb,src,dest);
end;


initialization
 MyDragMsg := RegisterWindowMessage('MYDRAGMESSAGE');




end.
